import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class DFT_Example extends JFrame{
	JTextArea textArea = new JTextArea(80, 60);
	JPanel controlPanel = new JPanel();
	JTextField sizeField = new JTextField(4);
	JComboBox modeBox;
	JButton goButton = new JButton("Go");

	private static String THETA = "\u0398"; // Unicode theta

	//Constructor
	public DFT_Example(){
		setTitle("DFT and inverses");
		
		String[] comboSelections = {"ramp", "high-pass ramp", "freq. shifted ramp"};
		modeBox = new JComboBox(comboSelections);
		
		controlPanel.add(new JLabel("size"));
		controlPanel.add(sizeField);
		controlPanel.add(modeBox);
		controlPanel.add(goButton);
		
		sizeField.setText("8");
		
		Container c = getContentPane();
		c.setLayout(new BorderLayout());	
		c.add(controlPanel, BorderLayout.NORTH);
		c.add(new JScrollPane(textArea), BorderLayout.CENTER);

		textArea.setFont(new Font("SansSerif", Font.PLAIN, 10));

		goButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e){
				goButtonPressed();
			}
		});

		setSize(500, 600);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	}

	private void goButtonPressed(){
		textArea.setText("");
		go();
	}

	// Major activity of the class
	public void go(){
		Complex inputSignal[];
		Complex transform[];
		Complex inverse[];
		int N = 16;
		int processingMode = 0;
		
		// read input size
		try{
			N = Integer.parseInt(sizeField.getText());
		}catch (Exception e){
			JOptionPane.showMessageDialog(this, "Bad size string",
				"DFT_Example", JOptionPane.WARNING_MESSAGE);
			return;
		}
		
		// read processing mode
		processingMode = modeBox.getSelectedIndex();		
		
		// Define input signal
		inputSignal = new Complex[N];
		if ( N ==1 ){
			textArea.append("(Note: For N = 1, I'm arbitrarily setting the function to = .5 ! )\n\n");
			inputSignal[0] = new Complex(0.5f, 0f);
			}
		else
			for (int i = 0; i < inputSignal.length; i++) 
				inputSignal[i] = new Complex((float)i/(float)inputSignal.length, 0);

		// Display input signal
		textArea.append("  Original Signal, N = " + N + "\n");
		textArea.append("  ------------------------------\n");
		append(textArea, "inputSignal", inputSignal);
	
		// DFT Transform
		transform = DFT.computeTransform(inputSignal);

		// Display transform
		textArea.append("  Transform:\n");
		textArea.append("  --------------\n");
		append(textArea, "transform", transform);

		// Perform spectral processing
		switch (processingMode){
			case 0:
				break;
			case 1:
				highPassFilter(transform);
				textArea.append("  Filtered transform:\n");
				textArea.append("  ------------\n");
				append(textArea, "filtered", transform);
				break;
			case 2:
				freqShift(transform);
				textArea.append("  Shifted transform (DC unmodified):\n");
				textArea.append("  ------------\n");
				append(textArea, "shifted", transform);
				break;		
		}		
		
		// DFT Inverse		
		inverse = DFT.computeInverse(transform);

		// Display inverse
		textArea.append("  Inverse:\n");
		textArea.append("  ------------\n");
		append(textArea, "inverse", inverse);

	}

	// high pass filter in spectral domain
	private void highPassFilter(Complex[] f){
		int i, N, Nby4;
		
		N = f.length;
		Nby4 = N/4;
		
		// first part of band
		for(i = 0; i<= Nby4; i++){
			f[i]=new Complex(0,0);		
		}
		// second part of band
		for(i=N-Nby4; i<N; i++){
			f[i]=new Complex(0,0);
		}
	}
	
	// frequency shift in spectral domain by N/4
	private void freqShift(Complex[] f){
		int i, N, Nby4;
		
		N = f.length;
		Nby4 = N/4;
		
		// first part of band
		for(i = 1; i<= Nby4; i++){
			f[i]=f[i+Nby4];
		}
		// second part of band
		for(i=N-Nby4; i<N; i++){
			f[i]=f[i-Nby4];
		}
		// middle (high freq) goes to zero
		for(i=Nby4+1; i<N-Nby4; i++){
			f[i] = new Complex(0,0);
		}
	}
	
	// Display utility method
	private void append(JTextArea jta, String name, Complex[] array){
		for (int i = 0; i < array.length; i++){
			jta.append(" " + name + "[" + i + "] = " + formatComplex(array[i]) + "\t");
			jta.append("r = " + formatReal(array[i].getMagnitude()) + "\t");
			jta.append(THETA + " = " + formatReal(array[i].getRadians()) +  "\t" +
				"deg" + " = " + formatReal(array[i].getDegrees()) +"\n");
		}
		jta.append("\n");
	}
	
	// Formatting methods for output
	public String formatReal(float x){
		x = Math.round(x*1000.0)/1000f;
		return new Float(x).toString();
	}	
	public String formatComplex(Complex x){
		float r = x.getReal();
		float i = x.getImag();
		
		// round for display
		r = Math.round(r*1000.0)/1000f;
		i = Math.round(i*1000.0)/1000f;
	
		// convert to string
		return new Float(r).toString() +
			(i>=0?" + ":" - ") +
			Math.abs(i) +
			"j";
	}

	// main
	public static void main(String[] args){
		DFT_Example x = new DFT_Example();
		x.show();
		x.go();
	}
}