/************************************************************
 * Glenn Josefiak                                           *
 * SW511                                                    *
 * Fairfield University                                     *
 * Fall 2002                                                *
 ************************************************************/

package audio;

import java.awt.event.*;
import javax.sound.sampled.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;
import dsp.*;

/************************************************************/
/** 
 * Extracts the MIDI note number corresponding to the pitch
 * of the sound at the PC's mic input.  Sends up pitch,
 * energy, and display events to report its results.
 */
public class AudioAnalyzer implements Runnable{
	public static final boolean DEBUG_FILTER = false;

	public static final int ENERGY_EVENT = 1;
	public static final int PITCH_EVENT = 2;
	public static final int REALTIME_MONITOR_EVENT = 3;
	public static final int PITCH_DISPLAY_EVENT = 4;

	private static final float SAMPLE_RATE_IN = 44100;
	private static final int DECIMATION_RATIO = 32;
	private static final float SAMPLE_RATE = SAMPLE_RATE_IN/DECIMATION_RATIO;

	private boolean running;
	private int amountRead;
	private int bufferSizeMS;
	private AudioFormat format;
	private int bufferSize;
	private byte buffer[];
	private short samples[];
	private KaiserFilter kaiserFilterDesigner;
	private FIRFilter filter;
	private PitchDetector pitchDetector;
	private TargetDataLine line;
	private SourceDataLine line1;
	private Thread inputThread;
	private EventListenerList listenerList = new EventListenerList();

	/********************************************************/
    /** 
	 * Constructs a new AudioAnalyzer
	 */
	public AudioAnalyzer(){
		float filterCoeffs[];
		int order;
	
		bufferSizeMS = 20; // default buffer duration
	
		kaiserFilterDesigner = new KaiserFilter();
		kaiserFilterDesigner.setFilterType("BP");  // "bandpass"
		kaiserFilterDesigner.setFreq1(70);         // Hz
		kaiserFilterDesigner.setFreq2(555);        // Hz
		kaiserFilterDesigner.setAtten(15);         // dB
		kaiserFilterDesigner.setTrBand(50);        // Hz
		kaiserFilterDesigner.setRate(SAMPLE_RATE_IN); // Hz
		
		order = kaiserFilterDesigner.estimatedOrder();
		kaiserFilterDesigner.design();

		filterCoeffs = new float[order];
		for (int i = 0; i < order; i++){
			filterCoeffs[i] = kaiserFilterDesigner.getCoeff(i);
		}

		if (DEBUG_FILTER){
			try{
				DataOutputStream dos = new DataOutputStream(new FileOutputStream("coeffs.txt"));
				for (int i = 0; i < order; i++){
					dos.writeBytes((new Float(filterCoeffs[i]).toString()) + "\r\n");
				}
			}catch (Exception e){
				System.out.println("Coefficient file write error.");
			}
		}
		
		filter = new FIRFilter(filterCoeffs);
		pitchDetector = new PitchDetector();
		pitchDetector.setSampleRate(SAMPLE_RATE);
	}

	/********************************************************/
    /** 
 	 * sets frame size in milliseconds
	 */
	public void setFrameSize(int ms){
		bufferSizeMS = ms;
	}

	/********************************************************/
    /** 
	 * gets frame size in milliseconds
	 */
	public int getFrameSize(){
		return bufferSizeMS;
	}

	/********************************************************/
    /** 
	 * gets latest note correlations
	 */
	public float[] getCorrelations(){
		return pitchDetector.getCorrelations();
	}

	/********************************************************/
    /** 
	 * starts the analyzer thread
	 */
	public void start(){
		inputThread = new Thread(this);
		inputThread.start();
	}
	
	/********************************************************/
    /** 
	 * stops the analyzer thread
	 */
	public void stop(){
		running = false;
	}

	/********************************************************/
    /** 
	 * The meat of the analyzer thread.
	 */
	public void run(){
		int pitch, energy;
	
		running = true;
		
		format = new AudioFormat(SAMPLE_RATE_IN, 16, 2, true, true);
		bufferSize = (int)(SAMPLE_RATE_IN * (bufferSizeMS/1000.0)) * format.getFrameSize();
		buffer = new byte[bufferSize];
		
		try{
			DataLine.Info info = new DataLine.Info(TargetDataLine.class, format);
			line = (TargetDataLine)AudioSystem.getLine(info);
			line.open(format);
			line.start();
		}catch(LineUnavailableException e){
			System.out.println(" Can't get an input line in AudioAnalyzer: " + e);
			return;
		}

    	// when 'running' is set to false, the thread exits.
		
		int monitorCount = 0; // how many buffers have been processed.
		while (running){
		
			Date startTime = new Date();
		
			// read data from microphone
			amountRead = line.read(buffer, 0, buffer.length);
							
			// extract a single channel of 16-bit linear PCM
			samples = extractMono(buffer, amountRead);

			// band pass filter input data
			filter.compute(samples);

			// decimate input data
			
			samples = decimate(samples, DECIMATION_RATIO);

			// compute pitch and notify listeners
			pitch = (int)pitchDetector.compute(samples);
			fireActionEvent(PITCH_EVENT, pitch);
			
			if ((monitorCount%5) == 0){
				fireActionEvent(PITCH_DISPLAY_EVENT, 0);
			}
			
			// compute energy and notify listeners
			energy = EnergyDetector.compute(samples);
			fireActionEvent(ENERGY_EVENT, energy);

			// post realtime monitor events
			
			if ((monitorCount%15) == 0){
				Date stopTime = new Date();
				
				// Calculate the number of milliseconds it took to
				// go through all the processing for this buffer of
				// audio.
				long t = stopTime.getTime()-startTime.getTime();
				// Calculate the processing time as a fraction of the
				// buffer length, i.e. how close we are to breaking
				// real time.
				float percent = (float)t/(float)bufferSizeMS*100;
				
				fireActionEvent(REALTIME_MONITOR_EVENT, (int)percent);
			}
			monitorCount++;
		}

		line.stop();
		line.close();
	}
	
	/********************************************************/
    /** 
     *	Get a single channel of mono data.
     *	Assumes input is 2-channel.  Ugly, but....
     */
	private short[] extractMono(byte[] buffer, int valid){
		short[] monoData = new short[valid/4];
		for (int i=0; i<valid; i+=4){
			monoData[i/4] = (short)(
				((int)buffer[i] << 8) + ((int)buffer[i+1] & 0xFF)
			);
		}		
		return monoData;
	}

	/********************************************************/
    /** 
     *	Reduce the sample rate by decimation.
     */
	private short[] decimate(short buffer[], int ratio){
		short output[] = new short[buffer.length/ratio];
		int j=0;
		for (int i=0; i<buffer.length/ratio; i++,j+=ratio){
			output[i] = buffer[j];
		}
		return output;
	}	
	
	/********************************************************/
    /** 
	 * allows other classes to listen for analyzer events
	 */
	public void addActionListener(ActionListener o){
		listenerList.add(ActionListener.class, o);
	}
	
	/********************************************************/
    /** 
	 * allows other classes to stop listening to analyzer events
	 */
	public void removeActionListener(ActionListener o){
		listenerList.remove(ActionListener.class, o);
	}

	/********************************************************/
    /** 
 	 * internally used for notifying listeners of events
	 */
	private void fireActionEvent(int type, int content){
		Object listeners[] = listenerList.getListeners(ActionListener.class);
		
		for (int i=0; i< listeners.length; i++){
			((ActionListener)listeners[i]).actionPerformed(
				new ActionEvent(
					this, 
					type,
					new Integer(content).toString()
				)
			);
		}
	}
	
}