/************************************************************
 * Glenn Josefiak                                           *
 * SW511                                                    *
 * Fairfield University                                     *
 * Fall 2002                                                *
 ************************************************************/

package gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.sound.midi.*;
//
import javax.sound.sampled.*;
import java.util.*;
import audio.*;

/************************************************************/
/** 
 * Displays a GUI for controlling a vocal instrument.  Uses
 * an AudioAnalyzer to extract the MIDI note number corresponding
 * to the pitch of the sound at the PC's mic input.  Controls
 * the MIDI synthesizer.
 */
public class VocalInstrument extends JFrame{

	private static final int FRAME_SIZE_MS = 100;
	private static final int DEFAULT_INSTRUMENT = 57;
	
	private JButton btnGo;
	private JButton btnStop;
	private JLabel  lblIndicator;
	private JPanel  pnlVerticalControls;
	private JLabel  lblRealtime;
	private JPanel  pnlRealtimeMonitor;
	private JTextField txtInstrument;
	private JButton btnInstrument;
	private JPanel  pnlInstrument;
	private JPanel  pnlSensitivity;
	private JSlider sliSensitivity;
	private JPanel  pnlIndicators;
	private NoteDisplay ndsGraph;
	private Container content;
	private AudioAnalyzer analyzer;
	
	private Synthesizer synthesizer;
	private MidiChannel channel;
	
	private int threshold = 500;
	private boolean onLastTime = false;
	private float currentPitch = 440.0f;
	private int currentNote;
	private int lastNote;
	
	/********************************************************/
   	/** 
	 * The constructor sets up the GUI and opens a MIDI channel
	 */
	public VocalInstrument(){
		super("Vocal Instrument: Voice to MIDI");
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		// Set up the GUI
		
		content = getContentPane();
		content.setLayout(new FlowLayout());
		
		pnlVerticalControls = new JPanel();
		pnlVerticalControls.setLayout(new GridLayout(3, 1, 6, 6));
		content.add(pnlVerticalControls);
		
		btnGo = new JButton("Go");
		pnlVerticalControls.add(btnGo);

		btnStop = new JButton("Stop");
		btnStop.setEnabled(false);
		pnlVerticalControls.add(btnStop);

		pnlRealtimeMonitor = new JPanel();
		pnlRealtimeMonitor.setLayout(new GridLayout(2, 1, 6, 6));
		pnlRealtimeMonitor.setBorder(new EmptyBorder(5,5,5,5));
		pnlRealtimeMonitor.setBackground(Color.white);
		pnlRealtimeMonitor.add(new JLabel("Realtime %"));
		lblRealtime = new JLabel("0");
		pnlRealtimeMonitor.add(lblRealtime);
		pnlVerticalControls.add(pnlRealtimeMonitor);
		
		pnlInstrument = new JPanel();
		pnlInstrument.setLayout(new GridLayout(3,1,6,6));
		pnlInstrument.setBorder(new EmptyBorder(5,5,5,5));
		pnlInstrument.setBackground(Color.white);
		pnlInstrument.add(new JLabel("Instrument"));
		txtInstrument = new JTextField(5);
		txtInstrument.setText(new Integer(DEFAULT_INSTRUMENT).toString());
		pnlInstrument.add(txtInstrument);
		btnInstrument = new JButton("Set");
		pnlInstrument.add(btnInstrument);
		content.add(pnlInstrument);

		pnlSensitivity = new JPanel();
		pnlSensitivity.setLayout(new BorderLayout(6,6));
		pnlSensitivity.setBorder(new EmptyBorder(5,5,5,5));
		pnlSensitivity.setBackground(Color.white);
		sliSensitivity = new JSlider(JSlider.VERTICAL, 25, 2000, 500);
		Hashtable hshSensTable = new Hashtable();
		hshSensTable.put(new Integer(25), new JLabel("25"));
		hshSensTable.put(new Integer(500), new JLabel("500"));
		hshSensTable.put(new Integer(1000), new JLabel("1000"));
		hshSensTable.put(new Integer(1500), new JLabel("1500"));
		hshSensTable.put(new Integer(2000), new JLabel("2000"));
		sliSensitivity.setLabelTable(hshSensTable);
		sliSensitivity.setPaintLabels(true);
		sliSensitivity.setOpaque(false);
		pnlSensitivity.add(new JLabel("Threshold"), BorderLayout.NORTH);
		pnlSensitivity.add(sliSensitivity, BorderLayout.CENTER);
		content.add(pnlSensitivity);

		pnlIndicators = new JPanel();
		pnlIndicators.setLayout(new BorderLayout(6,6));
		pnlIndicators.setBorder(new EmptyBorder(5,5,5,5));
		pnlIndicators.setBackground(Color.white);
		pnlIndicators.add(new JLabel("Energy Indicators"), BorderLayout.NORTH);
		ndsGraph = new NoteDisplay();
		pnlIndicators.add(ndsGraph, BorderLayout.CENTER);
		lblIndicator = new JLabel("0", JLabel.CENTER);
		lblIndicator.setOpaque(true);
		lblIndicator.setForeground(Color.red);
		lblIndicator.setBackground(Color.black);
		pnlIndicators.add(lblIndicator, BorderLayout.SOUTH);
		content.add(pnlIndicators);

		pack();

		// Set up the GUI listeners
		
		sliSensitivity.addChangeListener(new ChangeListener(){
			public void stateChanged(ChangeEvent e){
				sliderMoved();
			}
		});
		btnGo.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e){
				goButtonPressed();
			}
		});
		btnStop.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e){
				stopButtonPressed();
			}
		});
		btnInstrument.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e){
				instrumentNumberChanged();
			}
		});

		// Set up the audio elements
			
		analyzer = new AudioAnalyzer();
		analyzer.setFrameSize(FRAME_SIZE_MS);
		analyzer.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e){
				analyzerEventOccurred(e);
			}
		});
		
		try{
			synthesizer = MidiSystem.getSynthesizer();
			synthesizer.open();
			channel = synthesizer.getChannels()[0];
			channel.programChange(DEFAULT_INSTRUMENT);
		}catch(MidiUnavailableException x){
		}
	}

	/********************************************************/
    /** 
	 * Slider event handler
	 */
	private void sliderMoved(){
        if (!sliSensitivity.getValueIsAdjusting()) {
		   threshold = sliSensitivity.getValue();
        }
	}

	/********************************************************/
    /** 
	 * "Go" button event handler
	 */
	private void goButtonPressed(){
		btnGo.setEnabled(false);
		btnStop.setEnabled(true);
		analyzer.start();
	}
	
	/********************************************************/
    /** 
	 * "Stop" button event handler
	 */
	private void stopButtonPressed(){
		btnGo.setEnabled(true);
		btnStop.setEnabled(false);
		analyzer.stop();
		channel.allNotesOff();
	}

	/********************************************************/
    /** 
	 * Instrument change button event handler
	 */
	private void instrumentNumberChanged(){
		try{
			// Get text entered by the user
			int inst = Integer.parseInt(txtInstrument.getText());
			// MIDI instrument numbers are from 1 to 128; check range.
			if (inst < 1 || inst > 128){
				throw new Exception();
			}
			// Java uses instrument numbers 0 to 127.
			channel.programChange(inst-1);
		}catch (Exception e){
			JOptionPane.showMessageDialog(this, "Invalid Instrument Number",
			"Vocal Instrument", JOptionPane.ERROR_MESSAGE);
		}		
	}

	/********************************************************/
    /** 
	 * Analyzer event handler
	 */
	private void analyzerEventOccurred(ActionEvent e){		
		int level = 0;

		switch(e.getID()){			
			case AudioAnalyzer.ENERGY_EVENT:
				lblIndicator.setText(e.getActionCommand());
				level = Integer.parseInt(e.getActionCommand(), 10);
				if (level > threshold){
					lblIndicator.setBackground(Color.green);
					if (!onLastTime){
						channel.noteOn(currentNote,64);
						lastNote = currentNote;
						onLastTime = true;
					}
					else if (currentNote != lastNote){
						channel.noteOff(lastNote,64);
						channel.noteOn(currentNote,64);
						lastNote = currentNote;
					}
				}else{
					lblIndicator.setBackground(Color.black);
					if (onLastTime){
						channel.noteOff(lastNote,64);
						onLastTime = false;
					}
				}
				break;
			case AudioAnalyzer.PITCH_EVENT:
				currentNote = Integer.parseInt(e.getActionCommand());
				break;
			case AudioAnalyzer.PITCH_DISPLAY_EVENT:
				ndsGraph.setData(analyzer.getCorrelations());
				break;
			case AudioAnalyzer.REALTIME_MONITOR_EVENT:
				lblRealtime.setText(e.getActionCommand());
				break;
		}
	}

 	/********************************************************/
    /** 
	 * main entry point
	 */
	public static void main(String args[]){
		VocalInstrument v = new VocalInstrument();
		v.setLocation(100,100);
		v.show();
	}
}