package audio;

import futils.Futil;
import javax.sound.sampled.*;
import java.io.*;

public class AudioLogic {
    private boolean isCapturing = false;
    private boolean isPlaying = false;
    private AudioFormat format = defaultFormat;
    protected byte audioByteArray[];
    private String audioDescription;
    
    private static final AudioFormat defaultFormat =
    			new AudioFormat(16000, // sample rate
                                16,    // sampleSizeInBits
                                1,     // channels
                                true,  // signed
                                true); // big endian 
	
	/********************************************************/
	/**
	 * Play audio in the internal buffer.  This is a time-consuming
	 * operation.  The calling program is responsible for running this
	 * operation in a separate thread.
	 */
    public void playAudio() {
        int bufferSize;
        byte buffer[];
        AudioInputStream ais;
        SourceDataLine line = null;
    	
		// As play and record share both share the same audio data array, 
		// it is not possible to play and record at the same time.
		if (isCapturing){
			System.out.println(" Simultaneous play and record not allowed.");
			return;
		}

		if (audioByteArray == null){
            System.out.println(" No audio to play.");
            return;
        }

        InputStream is = new ByteArrayInputStream(audioByteArray);
        ais = new AudioInputStream(
                is,
                format,
                audioByteArray.length / format.getFrameSize()
        );

        // debug print
	    System.out.println(" Playing " + format);
        System.out.println(" Audio length = " + audioByteArray.length + " bytes");

        // set buffer size to 10ms for low latency
        bufferSize = ((int)(0.01 * format.getSampleRate())) * format.getFrameSize();
        buffer = new byte[bufferSize];

		isPlaying = true;
        try {
            DataLine.Info info = new DataLine.Info(
                   SourceDataLine.class, format);
            line = (SourceDataLine) AudioSystem.getLine(info);
            line.open(format);
            line.start();

            while (isPlaying){
            	int count = ais.read(buffer, 0, buffer.length);
            	if (count < 0)
            		isPlaying = false;
            	else if (count > 0)
            		line.write(buffer, 0, count);
            }

        } catch (IllegalArgumentException e){
        	System.out.println(" Unplayable audio format \n" + e);
        } catch (LineUnavailableException e) {
            System.out.println(" Can't get an output line in playAudio: " + e);
        } catch (IOException e) {
            System.out.println(" Audio output problem " + e);
        } finally {
        	if (line != null){
	            line.drain();
	            line.close();
        	}
        	isPlaying = false;
        }
    }


	/********************************************************/
	/**
	 * Record audio into the internal buffer.  This is a time-consuming
	 * operation.  The calling program is responsible for running this
	 * operation in a separate thread.
	 */
	public void captureAudio() {
       	int bufferSize;
	    byte buffer[];
	    TargetDataLine line = null;

		// As this class only maintains one array of recorded data,
		// multiple instances of capture cannot be occuring simultaneously.
		// Play and record share both share this same array, so it is also
		// not possible to play and record at the same time.
		if (isCapturing){
			System.out.println(" Multiple simultaneous captures not allowed.");
			return;
		}
		if (isPlaying){
			System.out.println(" Simultaneous play and record not allowed.");
			return;
		}		
		
	    format = defaultFormat;
	    System.out.println(" Recording " + format);

        // set buffer size to 10ms for low latency
        bufferSize = ((int)(0.01 * format.getSampleRate())) * format.getFrameSize();
	    buffer = new byte[bufferSize];
	    
	    ByteArrayOutputStream recordedByteStream = new ByteArrayOutputStream();
	    isCapturing = true;
	    try {
	        DataLine.Info info = new DataLine.Info(
	        	       	TargetDataLine.class, format);
	        line = (TargetDataLine) AudioSystem.getLine(info);
	        line.open(format);
	        line.start();
	        
	        while (isCapturing) {
	            int count = line.read(buffer, 0, buffer.length);
	            if (count > 0) {
	                recordedByteStream.write(
	                    buffer, 0, count);
	            }
	        }
	        recordedByteStream.close();
	        audioByteArray = recordedByteStream.toByteArray();
	        audioDescription = "Captured audio";
	        
	    } catch (LineUnavailableException e) {
	        System.out.println(" Can't get an input line in captureAudio: " + e);
	    } catch (IOException e) {
	        System.out.println(" Audio input problem " + e);
	    } finally {
	    	if (line != null){
		        line.stop();
		        line.close();
		    }
		    isCapturing = false;
	    }
	}
    
	/********************************************************/
	/**
	 * Save audio in the internal buffer to a file.
	 */
    public void saveAudioFile() {
		String writeFileName = null;

        if (isCapturing || isPlaying){
        	System.out.println(" Please stop audio before saving.");
        	return;
        }
        
		if (audioByteArray == null){
			System.out.println(" Nothing to save.");
			return;
		}        	
	    
        AudioInputStream ais = new AudioInputStream(
                new ByteArrayInputStream(audioByteArray),
                format,
                audioByteArray.length / format.getFrameSize()
        );

	    try {
	    	writeFileName = Futil.getWriteFileName();
	        FileOutputStream fos = new FileOutputStream(writeFileName);

	    	AudioSystem.write(ais, AudioFileFormat.Type.AU, fos);

	        fos.close();
	        audioDescription = writeFileName;
	        
	    } catch (IOException ioe) {
	        System.out.println(" No file created");
	    }
    }

	/********************************************************/
	/**
	 * Read audio from a file to the internal buffer.
	 */
    public void openAudioFile() {
		File readFile = null;
		
        if (isCapturing || isPlaying){
        	System.out.println(" Please stop audio before opening new file.");
        	return;
        }

		try{
			readFile = Futil.getReadFile();
			AudioInputStream ais = AudioSystem.getAudioInputStream(readFile);

			audioByteArray = new byte[ais.available()];
			ais.read(audioByteArray, 0, ais.available());
 		 	format = ais.getFormat();
 		 	
 		 	System.out.println(" Successfully read " + readFile.getName());
 		 	audioDescription = readFile.getName();
        } catch (UnsupportedAudioFileException a){
        	System.out.println(" Cannot read " + readFile.getName());
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

	/********************************************************/
	/**
	 * Return an AudioFormat object specifying the type of
	 * audio dealt with by this class.
	 */
    public AudioFormat getFormat() {
        return format;
    }

	/********************************************************/
	/**
	 * Return a String describing the audio currently stored
	 * in the internal buffer (file name or other source).
	 */
	public String getAudioDescription(){
		return audioDescription;
	}

	/********************************************************/
	/**
	 * Stop recording.
	 */
    public void stopCapturing() {
        isCapturing = false;
    }

	/********************************************************/
	/**
	 * Stop playing.
	 */
    public void stopPlaying() {
        isPlaying = false;
    }
}